/*  Author: Mobeus
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */
 
/* compile with valac --pkg gtk+-2.0 --pkg gmodule-2.0 Firewall_Genie.gs

    strip -s ./Firewall_Genie
    
    Memo: --pkg gmodule-2.0 is required for gtkbuilder and glade GUI file

    Support files are expected to be in the /usr/local/firewall-genie directory:
    firewall.glade
    firewall_template
    firewall_genie_help.html
*/

[indent=4]

uses
    Gtk

/* ****************************************/
/*   Global variable section              */
/******************************************/

firewall : string
ismodified : bool
supportdir : string
userdef : string

builder : Builder
savebtn : Button
vboxPrg : VBox
txtviewProg : TextView
txtBuf : TextBuffer
dialog_test : MessageDialog
window : Window

/* ****************************************/
/*   Glade / gtkbuilder widget callbacks  */
/******************************************/

def on_btnQuit_clicked (source : Button) : bool
    msg : string
    response : ResponseType
    msg = "The new firewall configuration has not been saved. Save it now?"
    if ismodified == true
        response = msgbox (window, MessageType.QUESTION, ButtonsType.YES_NO, msg)
        if response is ResponseType.YES
            on_btnSave_clicked ()
            return false
    Gtk.main_quit()
    return false

def on_btnDefaults_clicked (source : Button)
    fp0 : FileStream
    fp0 = FileStream.open (supportdir + "/FirewallGenie.rc", "w") //erase file
    config_gui ()
    install_firewall ()
    test_firewall ()

def on_btnSave_clicked ()
    var pStr = ""
    savebtn.set_sensitive (false)
    vboxPrg.visible = true
    config_save ()
    pStr = getAllowed ()
    pStr += " " + userdef
    install_firewall (pStr)
    test_firewall ()

def on_btnHelp_clicked (source : Button)
    cmd : string
    cmd = "defaultbrowser " + supportdir + "/firewall_genie_help.html"
    exec_async (cmd)

def modified ()
    /* called by all glade widgets that modify the allowed ports/ip setting */
    ismodified = true
    savebtn.set_sensitive (true)

/* **************************************/
/*   General routines                   */
/****************************************/

def config_save ()
    pStr : string
    udStr : string
    fp0 : FileStream    
    
    fp0 = FileStream.open (supportdir + "/FirewallGenie.rc", "w")
    pStr = getAllowed ()
    pStr = "[commonports] " + pStr
    udStr = "[userdefined]|" + userdef
    fp0.puts (pStr + "\n")
    fp0.puts (udStr)
    fp0.flush()

def config_gui () : bool
    fp0 : FileStream    
    cf : string
    lines : string
    length : size_t

    var chkbtn1 = builder.get_object ("chkFtp") as CheckButton
    var chkbtn2 = builder.get_object ("chkSsh") as CheckButton
    var chkbtn3 = builder.get_object ("chkHttp") as CheckButton
    var chkbtn4 = builder.get_object ("chkHttps") as CheckButton
    var entry = builder.get_object ("portlist") as Entry
    chkbtn1.set_active(false)
    chkbtn2.set_active(false)
    chkbtn3.set_active(false)
    chkbtn4.set_active(false)
    entry.set_text ("")

    cf = supportdir + "/FirewallGenie.rc"
    fp0 = FileStream.open (cf, "r")
    
    if fp0 is null
        fp0 = FileStream.open (cf, "w")
        return false
    try
        FileUtils.get_contents (cf, out lines, out length)
    except ex : GLib.Error
        print ("%s", "Could not read config file! --" + ex.message)
        
    var rclist = lines.split("\n")
        for line in rclist
            if line.contains("[commonports] ")
                var seg = line.split(" ")
                for port in seg
                    port = port.strip()
                    if port == "21/tcp"
                        chkbtn1.set_active(true)
                    if port == "22/tcp"
                        chkbtn2.set_active(true)
                    if port == "80/tcp"
                        chkbtn3.set_active(true)
                    if port == "443/tcp"
                        chkbtn4.set_active(true)

            if line.contains("[userdefined]|")
                var seg = line.split("|")
                entry.set_text (seg[1])

    ismodified = false
    return true

def exec (command : string, out status : int = 0) : string
    output : string
    errStr : string
    try
        Process.spawn_command_line_sync (command, out output, out errStr, out status)
    except ex : GLib.Error
        print ("%s", "spawn_command_line_sync failed! " + ex.message)
    return output

def exec_async (command : string) : bool
    try
        Process.spawn_command_line_async (command)
    except ex : GLib.Error
        print ("%s", "spawn_command_line_async failed! " + ex.message)
        return false
    return true

def install_firewall (permit : string = "") : bool
    cf : string
    ports : string
    str_line : string
    fp0 : FileStream ; fp1 : FileStream
        
    ismodified = false
    txtBuf.text = " "
    if permit == ""
        ports = "PERMIT=" +"\"\"" + "\n"
    else
        ports = "PERMIT=\"" + permit.strip () + "\"" + "\n"
        
    cf = supportdir + "/rc.firewall_template"
    fp0 = FileStream.open (cf, "r")
    cf = ""
    cf = supportdir + "/firewall_temp"
    fp1 = FileStream.open (cf, "w")

    var line = new array of char[128]
    while fp0.gets(line) != null
        str_line = (string)line; str_line.strip()
        if str_line.has_prefix("PERMIT=\"\"")
            str_line = ports.strip ()
        fp1.puts (str_line)
    
    fp1.flush()
    FileUtils.chmod (cf, 0750)
    return true
    
def test_firewall (permit : string = "") : bool
    cf : string
    response : ResponseType
    msg : string

    cf = supportdir + "/firewall_temp"

    msg = "Your firewall is now ready to be installed.  Would you like to verify your system configuration first?  This is a recommended step if this is the first time you are installing a firewall on this computer.  The testing procedure will not modify your current firewall configuration. Note that this operation may take some time depending on the speed of your computer."

    response = msgbox (window, MessageType.QUESTION, ButtonsType.YES_NO, msg)
    if response is ResponseType.YES
        append_output ("\n Testing the new firewall \n\n")
        exitstatus : int = 0
        msg = exec (cf, out exitstatus)
        append_output (msg)
        if exitstatus != 0
            msg = "Errors were detected in your system configuration. See the output section for specific details. A copy of the Linux Firewall initialization script preconfigured by this program is located at " + cf + ". No changes have been made to your existing firewall, and a new firewall has not been activated."
            response = msgbox (window, MessageType.ERROR, ButtonsType.OK, msg)
            return false /* firewall not installed or modified*/
    
    /* modify rc.local if necessary */
    fp0 : FileStream    
    lines : string
    length : size_t
    cf = "/etc/rc.d/rc.local"
    fp0 = FileStream.open (cf, "a")
    try
        FileUtils.get_contents (cf, out lines, out length)
    except ex : GLib.Error
        print ("%s", "Could not read " + cf + " -- " + ex.message)
    if lines.contains("/etc/rc.d/rc.firewall") == false
        exec ("cp " + cf + " " + cf + ".old")
        exec ("sync")
        msg = "if [ -x /etc/rc.d/rc.firewall ]; then\n" + "  /etc/rc.d/rc.firewall\n" + "fi"
        fp0.puts (msg)
        fp0.flush()

    /* stop firewall, backup firewall, copy new firewall to /etc/rc.d */
    append_output ("\n Stopping the current firewall. \n\n")
    append_output (exec(firewall + " stop"))
    append_output ("\n Backing up old firewall if it exists. \n")
    exec ("mv -f " + firewall + " " + firewall + ".old")
    exec ("sync")
    append_output ("\n Installing the new firewall. \n")
    cf = supportdir + "/firewall_temp"
    exec ("mv -f " + cf + " " + firewall)
    exec ("sync")
    FileUtils.chmod (firewall, 0750)
    append_output ("\n Starting the new firewall configuration. \n\n")
    append_output (exec(firewall + " start"))
    return true

def msgbox (parent : Window, type : MessageType, buttons : ButtonsType, text : string) : ResponseType
    dlg : MessageDialog
    dlg = new MessageDialog (parent, DialogFlags.DESTROY_WITH_PARENT | DialogFlags.MODAL, type, buttons, "")
    dlg.set_markup(text)
    var result = dlg.run()
    dlg.destroy()
    return (ResponseType) result
    
def append_output (txt : string)
    iter : TextIter
    mark : TextMark
    txtBuf.text += txt
    txtBuf.get_iter_at_line(out iter, -1) // -1 scrolls to end
    mark = txtBuf.create_mark("", iter, false)
    txtviewProg.scroll_mark_onscreen(mark)

def ischecked (chkBtn : string) : bool
    var chkbtn1 = builder.get_object (chkBtn) as CheckButton
    return chkbtn1.active

def getAllowed () : string
    allow : string; allow = ""
    if ischecked ("chkFtp") do allow = allow + "21/tcp "
    if ischecked ("chkSsh") do allow = allow + "22/tcp "
    if ischecked ("chkHttp") do allow = allow + "80/tcp "
    if ischecked ("chkHttps") do allow = allow + "443/tcp "
    var entry = builder.get_object ("portlist") as Entry
    userdef = entry.text
    allow = allow.strip ()
    return allow

/* **************************************/
/*   Program begins                     */
/****************************************/

init
    supportdir = "/usr/local/firewall-genie"
    firewall = "/etc/rc.d/rc.firewall"
    ismodified = false
    Gtk.init (ref args)
    builder = new Builder ()
    try
        builder.add_from_file (supportdir + "/firewall_genie.glade")
    except ex : GLib.Error
        print ("%s", "Could not load firewall.glade! -- " + ex.message)
    builder.connect_signals (null)
    window = builder.get_object ("window1") as Window
    window.show_all ()
    savebtn = builder.get_object ("btnSave") as Button
    savebtn.set_sensitive (false)
    vboxPrg = builder.get_object ("vboxPrg") as VBox
    txtviewProg = builder.get_object ("txtviewProg") as TextView
    txtBuf = txtviewProg.get_buffer ()
    config_gui ()
    Gtk.main ()

